from pathlib import Path

import ape.plugins

from ._secrets import Scope, get_secret_manager
from .accounts import KeyringAccount, KeyringAccountContainer
from .config import KeyringConfig


@ape.plugins.register(ape.plugins.Config)
def config_class():
    return KeyringConfig


@ape.plugins.register(ape.plugins.AccountPlugin)
def account_types():
    return KeyringAccountContainer, KeyringAccount


# Sync environment variables if configured to do so.
secret_manager = get_secret_manager(Path.cwd())
secret_manager.set_environment_variables()

__all__ = ["Scope", "secret_manager"]



from keyring.backend import KeyringBackend

from ape_keyring.storage import ACCOUNTS_TRACKER_KEY, SECRETS_TRACKER_KEY, SERVICE_NAME


class EphemeralBackend(KeyringBackend):
    """
    A keyring backend that requires Ape and is in-memory only,
    for testing purposes.
    """

    _storage = {ACCOUNTS_TRACKER_KEY: "", SECRETS_TRACKER_KEY: ""}

    @property
    def priority(cls):
        return 1

    def set_password(self, servicename, username, password):
        _require_ape(servicename, "Saving non-ape secret.")
        self._storage[username] = password

    def get_password(self, servicename, username):
        _require_ape(servicename, "Requesting non-ape secret.")

        if username in self._storage:
            return self._storage[username]

    def delete_password(self, servicename, username):
        _require_ape(servicename, "Deleting non-ape secret.")

        if username not in self._storage:
            raise AssertionError(f"Deleting non-stored username '{username}'.")

        del self._storage[username]


def _require_ape(service_name: str, msg: str):
    if service_name != SERVICE_NAME:
        raise ValueError(msg)



import pytest
from eth_account import Account
from eth_account.messages import encode_defunct


@pytest.fixture
def eip191_message():
    return encode_defunct(text="Hello Test")


def test_import(cli, runner, test_account, container, non_existing_alias):
    result = runner.invoke(
        cli, ["keyring", "accounts", "import", non_existing_alias], input=test_account.private_key
    )
    assert not result.exit_code, result.output
    assert "SUCCESS" in result.output
    assert test_account.address in result.output
    container.delete_account(non_existing_alias)


def test_list(cli, runner, keyring_account, existing_alias):
    result = runner.invoke(cli, ["keyring", "accounts", "list"])
    assert result.exit_code == 0, result.output
    assert existing_alias in result.output


def test_sign_message(keyring_account, runner, eip191_message):
    with runner.isolation("y\n"):
        signature = keyring_account.sign_message(eip191_message)

    signature_bytes = signature.encode_rsv()
    signer = Account.recover_message(eip191_message, signature=signature_bytes)
    assert signer == keyring_account.address


def test_set_autosign(keyring_account, eip191_message, runner):
    keyring_account.set_autosign(True)

    # The password is not required
    keyring_account.sign_message(eip191_message)

    keyring_account.set_autosign(False)

    # The password is now required
    with runner.isolation("y\n"):
        keyring_account.sign_message(eip191_message)


def test_delete(cli, runner, keyring_account):
    result = runner.invoke(cli, ["keyring", "accounts", "delete", keyring_account.alias])
    assert f"SUCCESS: Account '{keyring_account.alias}' removed from keying" in result.output

    # Verify account is NOT listed in 'list' command
    result = runner.invoke(cli, ["keyring", "accounts", "list"])
    assert keyring_account.alias not in result.output



import os

import pytest

from ape_keyring import Scope
from ape_keyring.config import KeyringConfig

GLOBAL_SECRET_KEY = "__GLOBAL_TEST_SECRET__"
PROJECT_SECRET_KEY = "__PROJECT_TEST_SECRET__"
GLOBAL_SECRET_VALUE = "test-global-secret-value"
PROJECT_SECRET_VALUE = "test-project-secret-value"

key_value_and_scope = pytest.mark.parametrize(
    "key,value,scope",
    [
        (GLOBAL_SECRET_KEY, GLOBAL_SECRET_VALUE, Scope.GLOBAL.value),
        (PROJECT_SECRET_KEY, PROJECT_SECRET_VALUE, Scope.PROJECT.value),
    ],
)


def clean(secret_manager):
    if GLOBAL_SECRET_KEY in secret_manager.global_keys:
        secret_manager.delete_secret(GLOBAL_SECRET_KEY)
    if PROJECT_SECRET_KEY in secret_manager.project_keys:
        secret_manager.delete_secret(PROJECT_SECRET_KEY)


@pytest.fixture(autouse=True)
def auto_clean(secret_manager):
    clean(secret_manager)
    yield
    clean(secret_manager)


@pytest.fixture
def temp_global_secret(secret_manager):
    if GLOBAL_SECRET_KEY not in secret_manager.global_keys:
        secret_manager.store_secret(GLOBAL_SECRET_KEY, GLOBAL_SECRET_VALUE, scope=Scope.GLOBAL)


@pytest.fixture
def temp_project_secret(secret_manager):
    if PROJECT_SECRET_KEY not in secret_manager.project_keys:
        secret_manager.store_secret(PROJECT_SECRET_KEY, PROJECT_SECRET_VALUE, scope=Scope.PROJECT)


@pytest.fixture
def temp_secrets(temp_global_secret, temp_project_secret):
    yield


@key_value_and_scope
def test_set(cli, runner, secret_manager, key, value, scope):
    result = runner.invoke(
        cli, ["keyring", "secrets", "set", key, "--scope", scope], input=f"{value}\n"
    )
    assert result.exit_code == 0, result.output


@pytest.mark.parametrize("key", (GLOBAL_SECRET_KEY, PROJECT_SECRET_KEY))
def test_list(cli, runner, temp_secrets, key):
    result = runner.invoke(cli, ["keyring", "secrets", "list"])
    assert key in result.output


@key_value_and_scope
def test_delete(cli, runner, temp_secrets, key, value, scope):
    _ = value
    result = runner.invoke(cli, ["keyring", "secrets", "delete", key, "--scope", scope])
    assert result.exit_code == 0, result.output


def test_config(config):
    plugin_config = config.get_config("keyring")
    assert isinstance(plugin_config, KeyringConfig)
    assert plugin_config.set_env_vars is True


@key_value_and_scope
def test_secrets_in_env(temp_secrets, secret_manager, key, value, scope):
    secret_manager.store_secret(key, value, scope=scope)
    assert os.environ.get(key) == value
    secret_manager.delete_secret(key, scope=scope)
    assert not os.environ.get(key)



from ape.exceptions import AccountsError, ApeException


class ApeKeyringException(ApeException):
    """
    A base error for the ape ``ape-keyring`` plugin.
    """


class ApeKeyringAccountError(ApeKeyringException, AccountsError):
    """
    A base error related to accounts in the ``ape-keyring`` plugin.
    """


class EmptyAliasError(ApeKeyringAccountError):
    def __init__(self):
        super().__init__("Alias cannot be empty.")


class SecretNotExistsError(ApeKeyringException):
    """
    Raised when trying to use a secret that does not exist.
    """

    def __init__(self, env_var_key: str):
        super().__init__(f"Secret '{env_var_key}' does not exist.")


class MissingSecretError(ApeKeyringException):
    """
    Raised when a secret was deleted outside of this plugin.
    """

    def __init__(self, alias: str):
        super().__init__(f"Missing secret for account with alias '{alias}'.")



import click
from ape import accounts
from ape.cli import ape_cli_context, existing_alias_argument, non_existing_alias_argument

from ape_keyring.utils import get_eth_account


@click.group("accounts")
def account_cli():
    """Manage accounts"""


@account_cli.command("list")
@ape_cli_context()
def _list(cli_ctx):
    """List accounts"""

    keyring_accounts = list(accounts.containers["keyring"].accounts)

    if not keyring_accounts:
        cli_ctx.logger.warning("No accounts found.")
        return

    num_accounts = len(keyring_accounts)
    click.echo(f"Found {num_accounts} account{'s' if num_accounts > 1 else ''}:")

    for account in keyring_accounts:
        alias_display = f" (alias: '{account.alias}')" if account.alias else ""
        click.echo(f"  {account.address}{alias_display}")


@account_cli.command(name="import")
@non_existing_alias_argument()
@ape_cli_context()
def _import(cli_ctx, alias):
    """Add a private key to keyring"""

    key = click.prompt("Enter the private key", hide_input=True)
    eth_account = get_eth_account(key)

    if not eth_account:
        cli_ctx.abort("Key could not be imported.")

    address = eth_account.address
    container = accounts.containers["keyring"]
    container.create_account(alias, key)
    cli_ctx.logger.success(f"A new account '{address}' has been added with the ID '{alias}'.")


@account_cli.command()
@ape_cli_context()
@existing_alias_argument()
def delete(cli_ctx, alias):
    """Remove a private key from keyring"""

    container = accounts.containers["keyring"]
    container.delete_account(alias)
    cli_ctx.logger.success(f"Account '{alias}' removed from keying.")


@account_cli.command()
@ape_cli_context()
def delete_all(cli_ctx):
    """Delete all private keys from keyring"""

    container = accounts.containers["keyring"]
    container.delete_all()
    cli_ctx.logger.success("Deleted all keyring accounts.")



from typing import Iterator, Optional

from ape.api import AccountAPI, AccountContainerAPI, TransactionAPI
from ape.exceptions import AccountsError
from ape.logging import logger
from ape.types import AddressType, MessageSignature, TransactionSignature
from ape.utils import cached_property
from eth_account import Account as EthAccount  # type: ignore
from eth_account.messages import SignableMessage
from eth_utils import to_bytes

from ape_keyring.exceptions import EmptyAliasError, MissingSecretError
from ape_keyring.storage import SecretStorage, account_storage
from ape_keyring.utils import agree_to_sign, get_eth_account


class KeyringAccountContainer(AccountContainerAPI):
    storage: SecretStorage = account_storage

    @property
    def aliases(self) -> Iterator[str]:
        yield from [a for a in self.storage.keys if a]

    @property
    def accounts(self) -> Iterator[AccountAPI]:
        for alias in self.aliases:
            yield self.load(alias)

    def load(self, alias: str) -> "KeyringAccount":
        return KeyringAccount(
            storage=self.storage, storage_key=alias, container=self
        )  # type: ignore

    def __len__(self) -> int:
        return len([a for a in self.aliases if a])

    def create_account(self, alias: str, secret: str):
        if not alias:
            raise EmptyAliasError()

        self.storage.store_secret(alias, secret)
        if not self.storage.get_secret(alias):
            raise AccountsError(f"Failed to create account '{alias}'")

    def delete_account(self, alias: str):
        if not alias:
            raise EmptyAliasError()

        self.storage.delete_secret(alias)

    def delete_all(self):
        self.storage.delete_all()


class KeyringAccount(AccountAPI):
    storage: SecretStorage
    storage_key: str
    cached_address: Optional[AddressType] = None
    __autosign: bool = False

    def __repr__(self):
        return f"<{self.__class__.__name__} address={self.address} alias={self.alias}>"

    @property
    def alias(self) -> str:
        return self.storage_key

    @cached_property
    def address(self) -> AddressType:
        if not self.cached_address:
            eth_account = get_eth_account(self.__key)
            if eth_account:
                ethereum = self.network_manager.get_ecosystem("ethereum")
                self.cached_address = ethereum.decode_address(eth_account.address)

        if not self.cached_address:
            raise AccountsError("Account private key corrupted.")

        return self.cached_address

    @property
    def __key(self) -> str:
        key = self.storage.get_secret(self.storage_key)
        if not key:
            raise MissingSecretError(self.storage_key)

        return key

    def sign_message(self, msg: SignableMessage) -> Optional[MessageSignature]:
        if not self.__autosign and not agree_to_sign(msg, "message"):
            return None

        signed_msg = EthAccount.sign_message(msg, self.__key)
        return MessageSignature(
            v=signed_msg.v, r=to_bytes(signed_msg.r), s=to_bytes(signed_msg.s)
        )  # type: ignore

    def sign_transaction(self, txn: TransactionAPI) -> Optional[TransactionSignature]:
        if not self.__autosign and not agree_to_sign(txn, "transaction"):
            return None

        signed_txn = EthAccount.sign_transaction(txn.dict(), self.__key)
        return TransactionSignature(
            v=signed_txn.v, r=to_bytes(signed_txn.r), s=to_bytes(signed_txn.s)
        )  # type: ignore

    def set_autosign(self, enabled: bool):
        """
        Allow this account to automatically sign messages and transactions.

        Args:
            enabled (bool): ``True`` to enable, ``False`` to disable.
        """
        logger.warning("Danger! This account will now sign any transaction it's given.")
        self.__autosign = enabled
